#include "bitstream.h"
//#include "exception.h"

/*
 * Le but de ce fichier est de fournir des fonctions permettant
 * d'crire/lire un flot de bit, bit par bit dans ou  partir d'un fichier.
 * On ne mlange pas criture et lecture, on ouvre pour lire OU
 * pour crire, pas les 2.
 *
 * Le principe est simple on utilise un buffer d'entre/sortie.
 * On ne fait rellement la sortie que lorsque le buffer est plein
 * ou l'entre quand il est vide.
 */


/*
 * Cette structure contient toutes les informations
 * permettant d'ecrire (ou de lire) les bits un par un dans un fichier.
 * Evidemment aucune fonction de gestion de fichier ne permet de faire cela.
 * On va donc stocker les bits un par un dans un entier (buffer)
 * et quand celui-ci sera plein, on le stockera dans le fichier.
 *
 * Pour la lecture, le procd est inverse, on lit le buffer.
 * Puis on en extrait les bits un par un
 * jusqu' ce qu'il soit vide.
 */
struct bitstream
{
  FILE          *fichier ;		     /* En lecture ou Ecriture */
  Buffer_Bit     buffer ;		     /* Tampon intermediaire */
  Position_Bit   nb_bits_dans_buffer ;	     /* Nb bits dans le tampon */
  Booleen        ecriture ;		     /* Faux, si ouvert avec "r" */
} ;

/*
 * Cette fonction alloue la structure, l'initialise et ouvre le fichier.
 * Evidemment elle vide le buffer.
 * Le "mode" est passe tel quel  la fonction "fopen".
 *
 * On considre que le fichier est ouvert en lecture si
 * le mode est "r"
 *
 * Il faut que vous testiez un nom de fichier particulier : "-"
 *     - Ce fichier est l'entre standard s'il est ouvert en lecture
 *     - Ce fichier est la sortie standard s'il est ouvert en criture.
 *
 * Un fichier est ouvert en lecture ou ecriture.
 * Pas les deux en mme en temps.
 *     - En lecture on ne fera que des "getbit" (dfinie plus loin)
 *     - En criture on ne fera que des "putbit" (dfinie plus loin)
 *
 * Si le fichier ne peut tre ouvert, on lance l'exception :
 *         "Exception_fichier_ouverture"
 * Pour plus d'explications sur les exceptions, regardez "exception.h"
 */

struct bitstream *open_bitstream(const char *fichier, const char* mode)
{

  struct bitstream* s_bitstream ;
  ALLOUER(s_bitstream,1) ;
  s_bitstream->ecriture = strcmp(mode,"w") == 0 || strcmp(mode,"w+") == 0 ;

  s_bitstream->nb_bits_dans_buffer = 0 ;
  
  if(!strcmp(fichier,"-")){
  
    if(s_bitstream->ecriture)
      s_bitstream->fichier = stdout ;
    else
      s_bitstream->fichier = stdin ;

  }else{
    s_bitstream->fichier = fopen(fichier, mode) ;     
  }

  if(s_bitstream->fichier == NULL){
    free(s_bitstream) ;
    fprintf(stderr,"Erreur : open_bitstream\n") ;
    abort() ;

    //    EXCEPTION_LANCE(Exception_fichier_ouverture) ;
  }
  return s_bitstream ;
}

struct bitstream *open_bitstream_file(FILE *fichier, const char* mode)
{

  struct bitstream* s_bitstream ;
  ALLOUER(s_bitstream,1) ;
  s_bitstream->ecriture = strcmp(mode,"w") == 0 || strcmp(mode,"w+") == 0 ;

  s_bitstream->nb_bits_dans_buffer = 0 ;
  
  
  s_bitstream->fichier = fichier ;     
  

  if(s_bitstream->fichier == NULL){
    free(s_bitstream) ;
    fprintf(stderr,"Erreur : open_bitstream\n") ;
    abort() ;

    //    EXCEPTION_LANCE(Exception_fichier_ouverture) ;
  }
  return s_bitstream ;
}

/*
 * Cette fonction ne fait rien si le fichier est ouvert en lecture.
 * 
 * Si le buffer n'est pas vide :
 *    - Cette fonction stocke le buffer dans le fichier
 *      que le buffer soit "complet" ou non.
 *      Elle ne stocke rien si le buffer est vide
 *    - Elle vide ensuite le buffer.
 * Cette fonction n'est appele que lorsque le fichier est ouvert
 * en criture.
 *
 * Si il y a une erreur d'criture, elle lance l'exception :
 *         "Exception_fichier_ecriture"  
 */

void flush_bitstream(struct bitstream *b)
{
  if (b->ecriture){
    if(b->nb_bits_dans_buffer > 0){

      if(fputc(b->buffer,b->fichier) != b->buffer){
	fprintf(stderr,"Erreur : flush_bistreamt\n") ;
	abort() ;
	//	EXCEPTION_LANCE(Exception_fichier_ecriture) ;
      }else{
	//b->buffer = 0 ;
	b->nb_bits_dans_buffer = 0 ;
      }
    }
  }
}

/*
 * Avant de fermer le fichier ouvert en criture on copie le buffer
 * dans le fichier.
 * On ferme MEME si le fichier est l'entre ou la sorte standard.
 *
 * Si jamais, il y a une erreur de fermeture, on lance l'exception
 *         Exception_fichier_fermeture
 */

void close_bitstream(struct bitstream *b)
{
    flush_bitstream(b) ;
 
    if(fclose(b->fichier)){
      fprintf(stderr,"Erreur : close_bitstream\n") ;
      abort() ;
    }
    //    EXCEPTION_LANCE(Exception_fichier_fermeture) ;

  free(b) ;
}

void close_bitstream_file(struct bitstream *b)
{
    flush_bitstream(b) ;
    free(b) ;
}

/*
 * Cette fonction ajoute le "bit" dans le buffer.
 *    - Si celui-ci est plein, alors on sauvegarde le buffer
 *      dans le fichier avec "flush_bitstream".
 *
 *    - On pose le bit dans le buffer.
 *
 * Cette fonction n'est appele que lorsque
 * le fichier est ouvert en criture.
 *
 * On stocke les bits du numro le plus grand possible jusqu' 0.
 * On les stocke donc du poids fort au poid faible, c.a.d. de gauche  droite
 *
 * Si le fichier est ouvert en lecture, on lance l'exception
 *         Exception_fichier_ecriture_dans_fichier_ouvert_en_lecture
 */

void put_bit(struct bitstream *b, Booleen bit)
{
  if(b->ecriture ){

    if(b->nb_bits_dans_buffer >= NB_BITS)//sizeof(b->buffer)*8) 
      flush_bitstream(b) ;

    int position = (NB_BITS-1) - b->nb_bits_dans_buffer ;
    b->buffer = pose_bit(b->buffer, position, bit) ;
    b->nb_bits_dans_buffer++ ;

  }else{
    fprintf(stderr,"Erreur : put_bit\n") ;
    abort() ;
  }
  //else
  // EXCEPTION_LANCE(Exception_fichier_ecriture_dans_fichier_ouvert_en_lecture) ;

}


/*
 * Cette fonction lit un bit du buffer (du poid fort au poid faible)
 * Si le buffer est vide, elle va le lire dans le fichier.
 * Les valeurs retournes possibles sont :
 *    - (Faux)
 *    - (Vrai)
 *
 * Cette fonction n'est appele que lorsque
 * le fichier est ouvert en lecture.
 *
 * ATTENTION ATTENTION : fgetc() RETOURNE UN int, PAS UN char.
 * faites donc attention aux affectations et aux tests par rapport  EOF.
 * Posez-vous les questions :
 *    - Pourquoi la valeur de retour est "int" et non "char"
 *    - Quel est le type de "b->buffer"
 *
 * En cas d'erreur de lecture (fin de fichier) on lance l'exception
 *         Exception_fichier_lecture
 *
 * Si le fichier est ouvert en criture, on lance l'exception
 *         Exception_fichier_lecture_dans_fichier_ouvert_en_ecriture
 */

Booleen get_bit(struct bitstream *b)
{
  int result = 0 ;
  if (b->ecriture){
    fprintf(stderr,"Erreur : get_bit\n") ;
    abort() ;
    //    EXCEPTION_LANCE(Exception_fichier_lecture_dans_fichier_ouvert_en_ecriture) ;
  }else{

    if(b->nb_bits_dans_buffer == 0) {
	int first = fgetc(b->fichier) ;

	
	if(first == EOF){
	  fprintf(stderr,"Erreur : put_bit\n") ;
	  abort() ;
	}
	  //EXCEPTION_LANCE(Exception_fichier_lecture) ;
	
	b->buffer = first ;
	b->nb_bits_dans_buffer = NB_BITS ;
      }
    b->nb_bits_dans_buffer-- ;
    result = prend_bit(b->buffer,b->nb_bits_dans_buffer) ;
  }

  return result ;
}

/*
 * Ne modifiez pas la fonctions suivantes
 *
 * VOUS NE DEVEZ PAS LES UTILISER, ELLE SONT SEULEMENT LA
 * POUR QU'IL SOIT POSSIBLE D'ECRIRE DES TESTS.
 */
FILE *bitstream_get_file(const struct bitstream *b)
{
  return( b->fichier ) ;
}
Booleen bitstream_en_ecriture(const struct bitstream *b)
{
  return( b->ecriture ) ;
}
int bitstream_nb_bits_dans_buffer(const struct bitstream *b)
{
  return( b->nb_bits_dans_buffer ) ;
}
